# TODO: rocks db

from opteryx.managers.kvstores import BaseKeyValueStore


class MongoDbKVStore(BaseKeyValueStore):
    def __init__(self):
        raise NotImplementedError("Mongo KV Store not implemented")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from typing import List

from opteryx.exceptions import NotSupportedError
from opteryx.managers.expression import NodeType


PUSHABLE_OPERATORS = {
    # these are almost universally supported
    "Gt": ">",
    "Lt": "<",
    "Eq": "==",  # usually ==, sometimes =
    "NotEq": "!=",  # usually !=, sometimes <>
    "GtEq": ">=",
    "LtEq": "<=",
}


class PredicatePushable(abc.ABC):
    @staticmethod
    def to_dnf(root):
        """
        Convert a filter to the form used by the selection pushdown

        Version 1 only does simple predicate filters in the form
            (identifier, operator, literal)
        although we can form conjuntions (ANDs) by chaining them.

        Return None if we can't convert, or don't support the filter.
        """

        def _predicate_to_dnf(root):
            # Reduce look-ahead effort by using Exceptions to control flow
            if root.token_type == NodeType.AND:
                left = _predicate_to_dnf(root.left)
                right = _predicate_to_dnf(root.right)
                if not isinstance(left, list):
                    left = [left]
                if not isinstance(right, list):
                    right = [right]
                left.extend(right)
                return left
            if root.token_type != NodeType.COMPARISON_OPERATOR:
                raise NotSupportedError()
            if not root.value in PUSHABLE_OPERATORS:
                # not all operators are universally supported
                raise NotSupportedError()
            if root.left.token_type != NodeType.IDENTIFIER:
                raise NotSupportedError()
            if root.left.token_type in (
                NodeType.LITERAL_NUMERIC,
                NodeType.LITERAL_VARCHAR,
            ):
                # not all operands are universally supported
                raise NotSupportedError()
            return (root.left.value, PUSHABLE_OPERATORS[root.value], root.right.value)

        try:
            dnf = _predicate_to_dnf(root)
            if not isinstance(dnf, list):
                dnf = [dnf]
        except NotSupportedError:
            return None
        return dnf

    _predicates: List = []

    def push_predicate(self, predicate):
        """
        Push the predicate to the set - this creates a set of ANDs
        """
        dnfed = PredicatePushable.to_dnf(predicate)
        if dnfed is None:
            # we can't push all predicates everywhere
            return False
        self._predicates.extend(dnfed)
        return True



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from os import environ

import yaml

# python-dotenv allows us to create an environment file to store secrets. If
# there is no .env it will fail gracefully.
try:
    import dotenv  # type:ignore
except ImportError:  # pragma: no cover
    dotenv = None  # type:ignore

_env_path = Path(".") / ".env"
_config_values: dict = {}

#  deepcode ignore PythonSameEvalBinaryExpressiontrue: false +ve, values can be different
if _env_path.exists() and (dotenv is None):  # pragma: no cover  # nosemgrep
    # using a logger here will tie us in knots
    print("`.env` file exists but `dotEnv` not installed.")
elif dotenv is not None:  # pragma: no cover
    print("Loading environment variables from `.env`")
    dotenv.load_dotenv(dotenv_path=_env_path)

try:  # pragma: no cover
    _config_path = Path(".") / "opteryx.yaml"
    if _config_path.exists():
        with open(_config_path, "rb") as _config_file:
            _config_values = yaml.safe_load(_config_file)
        print(f"Loading config from {_config_path}")
except Exception as exception:  # pragma: no cover # it doesn't matter why - just use the defaults
    print(f"Config file {_config_path} not used - {exception}")


def get(key, default=None):
    value = environ.get(key)
    if value is None:
        value = _config_values.get(key, default)
    return value


# fmt:off

# These are 'protected' properties which cannot be overridden by a single query

# GCP project ID - for Google Cloud Data
GCP_PROJECT_ID: str = get("GCP_PROJECT_ID")
# Mapping prefixes to readers - the default is to use disk
DATASET_PREFIX_MAPPING: dict = get("DATASET_PREFIX_MAPPING", {"_":"disk"})
# Data Partitioning
PARTITION_SCHEME: str = get("PARTITION_SCHEME", None)
# The maximum number of evictions by a single query
MAX_CACHE_EVICTIONS: int = int(get("MAX_CACHE_EVICTIONS", 25))
# Maximum size for items saved to the buffer cache
MAX_SIZE_SINGLE_CACHE_ITEM: int = get("MAX_SIZE_SINGLE_CACHE_ITEM", 1024 * 1024)
# The local buffer pool size
LOCAL_BUFFER_POOL_SIZE: int = int(get("LOCAL_BUFFER_POOL_SIZE", 256))
# don't try to raise the priority of the server process
DISABLE_HIGH_PRIORITY: bool = bool(get("DISABLE_HIGH_PRIORITY", False))
# don't output resource (memory) utilization information
ENABLE_RESOURCE_LOGGING: bool = bool(get("ENABLE_RESOURCE_LOGGING", False))
# only push equals predicates
ONLY_PUSH_EQUALS_PREDICATES: bool = bool(get("ONLY_PUSH_EQUALS_PREDICATES", False))
# size of pages to push between steps
PAGE_SIZE: int = int(get("PAGE_SIZE", 64 * 1024 * 1024))
# fmt:on



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# faust-streaming-rocksdb

"""
This is opinionated for use as a metadata cache, it can be used as a reader cache
because they implement the same interface, but that would cache all of the files
locally, which is unlikely to be what is wanted.

"""
import io

from typing import Iterable

from opteryx.exceptions import MissingDependencyError
from opteryx.managers.kvstores import BaseKeyValueStore


ROCKS_DB = True

try:
    import rocksdb
except ImportError:
    ROCKS_DB = False


class RocksDB_KVStore(BaseKeyValueStore):
    def __init__(self, location):

        super().__init__(location)
        self._location += ".rocks"

        if not ROCKS_DB:
            raise MissingDependencyError(
                "`RocksDB` is missing, please install or include in requirements.txt"
            )
        self._db = rocksdb.DB(location, rocksdb.Options(create_if_missing=True))

    @staticmethod
    def can_use():
        return ROCKS_DB

    def get(self, key):
        if hasattr(key, "encode"):
            key = key.encode()
        byte_reponse = self._db.get(key)
        if byte_reponse:
            return io.BytesIO(byte_reponse)
        return None

    def set(self, key, value):
        if hasattr(key, "encode"):
            key = key.encode()
        value.seek(0, 0)
        self._db.put(key, value.read())
        value.seek(0, 0)

    def contains(self, keys: Iterable) -> Iterable:
        return [key for key in keys if self.get(key) is not None]



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Build Statistics Node

This is a SQL Query Execution Plan Node.

Gives information about a dataset's columns
"""
import datetime

from functools import reduce
from typing import Iterable

import numpy
import orjson
import pyarrow

from opteryx.attribute_types import OPTERYX_TYPES, determine_type
from opteryx.exceptions import SqlError
from opteryx.models import Columns, QueryProperties
from opteryx.operators import BasePlanNode

MAX_COLLECTOR: int = 8
MAX_VARCHAR_SIZE: int = 64  # long strings tend to lose meaning
MAX_DATA_SIZE: int = 100 * 1024 * 1024
UNIX_EPOCH = datetime.datetime(1970, 1, 1)


def _to_unix_epoch(date):
    if date.as_py() is None:
        return numpy.nan
    # Not all platforms can handle negative timestamp()
    # https://bugs.python.org/issue29097
    return (date.as_py() - UNIX_EPOCH).total_seconds()


def increment(dic: dict, value):
    if value in dic:
        dic[value] += 1
    else:
        dic[value] = 1


def _statitics_collector(pages):
    """
    Collect summary statistics about each column
    """
    from opteryx.third_party import distogram

    empty_profile = orjson.dumps(
        {
            "name": None,
            "type": [],
            "count": 0,
            "missing": 0,
            "bytes": 0,
            "most_frequent_values": None,
            "most_frequent_counts": None,
            "numeric_range": None,
            "varchar_range": None,
            "distogram_values": None,
            "distogram_counts": None,
        }
    )
    target_metadata = None

    for page in pages:

        uncollected_columns = []
        profile_collector: dict = {}

        columns = Columns(page)
        table_path = columns.table_path

        for block in page.to_batches(10000):

            for column in page.column_names:

                column_data = block.column(column)

                profile = profile_collector.get(column, orjson.loads(empty_profile))
                _type = determine_type(str(column_data.type))
                if _type not in profile["type"]:
                    profile["type"].append(_type)

                profile["count"] += len(column_data)
                profile["bytes"] += column_data.nbytes
                profile["missing"] += column_data.null_count

                # interim save
                profile_collector[column] = profile

                # don't collect problematic columns
                if column in uncollected_columns:
                    continue

                # to prevent problems, we set some limits
                if column_data.nbytes > MAX_DATA_SIZE:  # pragma: no cover
                    if column not in uncollected_columns:
                        uncollected_columns.append(column)
                    continue

                # don't collect columns we can't analyse
                if _type in (
                    OPTERYX_TYPES.LIST,
                    OPTERYX_TYPES.STRUCT,
                    OPTERYX_TYPES.OTHER,
                ):
                    continue

                # long strings are meaningless
                if _type == OPTERYX_TYPES.VARCHAR:

                    column_data = [v.as_py() for v in column_data if v.is_valid]

                    max_len = reduce(
                        lambda x, y: max(len(y), x),
                        column_data,
                        0,
                    )
                    if max_len > MAX_VARCHAR_SIZE:
                        if column not in uncollected_columns:
                            uncollected_columns.append(column)
                        continue

                    # collect the range values
                    if len(column_data) > 0:
                        varchar_range_min = min(column_data)
                        varchar_range_max = max(column_data)

                        if profile["varchar_range"] is not None:
                            varchar_range_min = min(
                                varchar_range_min, profile["varchar_range"][0]
                            )
                            varchar_range_max = max(
                                varchar_range_max, profile["varchar_range"][1]
                            )

                        profile["varchar_range"] = (
                            varchar_range_min,
                            varchar_range_max,
                        )

                # convert TIMESTAMP into a NUMERIC (seconds after Unix Epoch)
                if _type == OPTERYX_TYPES.TIMESTAMP:
                    column_data = (_to_unix_epoch(i) for i in column_data)
                elif _type != OPTERYX_TYPES.VARCHAR:
                    column_data = (i.as_py() for i in column_data)
                # remove empty values
                column_data = numpy.array(
                    [i for i in column_data if i not in (None, numpy.nan)]
                )

                if _type in (OPTERYX_TYPES.BOOLEAN):
                    # we can make it easier to collect booleans
                    counter = profile.get("counter")
                    if counter is None:
                        counter = {"True": 0, "False": 0}
                    trues = sum(column_data)
                    counter["True"] += trues
                    counter["False"] += column_data.size - trues
                    profile["counter"] = counter

                if _type == OPTERYX_TYPES.VARCHAR and profile.get("counter") != {}:
                    # counter is used to collect and count unique values
                    vals, counts = numpy.unique(column_data, return_counts=True)
                    counter = {}
                    if len(vals) <= MAX_COLLECTOR:
                        counter = dict(zip(vals, counts))
                        for k, v in profile.get("counter", {}).items():
                            counter[k] = counter.pop(k, 0) + v
                        if len(counter) > MAX_COLLECTOR:
                            counter = {}
                    profile["counter"] = counter

                if _type in (OPTERYX_TYPES.NUMERIC, OPTERYX_TYPES.TIMESTAMP):
                    # populate the distogram, this is used for distribution statistics
                    dgram = profile.get("dgram")
                    if dgram is None:
                        dgram = distogram.Distogram()  # type:ignore
                    dgram.bulkload(column_data)
                    profile["dgram"] = dgram

                profile_collector[column] = profile

        buffer = []

        for column, profile in profile_collector.items():
            profile["name"] = columns.get_preferred_name(column)
            profile["type"] = ", ".join(profile["type"])

            if column not in uncollected_columns:

                dgram = profile.pop("dgram", None)
                if dgram:
                    # force numeric types to be the same
                    profile["numeric_range"] = (
                        numpy.double(dgram.min),
                        numpy.double(dgram.max),
                    )
                    profile["distogram_values"], profile["distogram_counts"] = zip(
                        *dgram.bins
                    )
                    profile["distogram_values"] = numpy.array(
                        profile["distogram_values"], numpy.double
                    )

                counter = profile.pop("counter", None)
                if counter:
                    counts = list(counter.values())
                    if min(counts) != max(counts):
                        profile["most_frequent_values"] = [
                            str(k) for k in counter.keys()
                        ]
                        profile["most_frequent_counts"] = counts

            # remove collectors
            profile.pop("dgram", None)
            profile.pop("counter", None)

            buffer.append(profile)

        table = pyarrow.Table.from_pylist(buffer)

        if target_metadata is None:
            table = Columns.create_table_metadata(
                table=table,
                expected_rows=len(buffer),
                name="statistics",
                table_aliases=[],
                disposition="statistics",
                path=table_path,
            )
            target_metadata = Columns(table)
        else:
            table = target_metadata.apply(table, path=table_path)
        yield table


class BuildStatisticsNode(BasePlanNode):
    def __init__(self, properties: QueryProperties, **config):
        super().__init__(properties=properties)

    @property
    def name(self):  # pragma: no cover
        return "Analyze Table"

    @property
    def config(self):  # pragma: no cover
        return ""

    def execute(self) -> Iterable:

        if len(self._producers) != 1:
            raise SqlError(f"{self.name} on expects a single producer")

        data_pages = self._producers[0]  # type:ignore

        if data_pages is None:
            return None

        return _statitics_collector(data_pages.execute())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

FUNCTION_KIND_FUNCTION: int = 1
FUNCTION_KIND_AGGREGATOR: int = 2


class FunctionProperty(int, Enum):
    NAME: int = 0
    TYPE: int = 1
    DESCRIPTION: int = 2
    RETURN_TYPE: int = 3
    NULLABLE_RETURN: int = 4  # can it ever return null
    MINIMUM_RETURN: int = 5  # lowest value that can be returned
    MAXIMUM_RETURN: int = 6  # greatest value that can be returned
    PARAMETERS: int = 7
    RELATED: int = 8  # functions that are related
    COST: int = 9  # indicative time to execute function 1 million times


# fmt:off
FUNCTION_REGISTRY = [
    ("ALL", 2, "Returns `true` if all items in group are true", bool, None, None, None, None, None, None),
    ("ANY", 2, "Returns `true` if any items in a group are true", bool, None, None, None, None, None, None),
]

"""
    "APPROXIMATE_MEDIAN": "approximate_median",
    "COUNT": "count",  # counts only non nulls
    "COUNT_DISTINCT": "count_distinct",
    "DISTINCT": "distinct",
    "LIST": "hash_list",
    "MAX": "max",
    "MAXIMUM": "max",  # alias
    "MEAN": "mean",
    "AVG": "mean",  # alias
    "AVERAGE": "mean",  # alias
    "MIN": "min",
    "MINIMUM": "min",  # alias
    "MIN_MAX": "min_max",
    "ONE": "hash_one",
    "ANY_VALUE": "hash_one",
    "PRODUCT": "product",
    "STDDEV": "stddev",
    "SUM": "sum",
    "VARIANCE": "variance",
"""

# fmt:on
